const fs = require('fs');
const { resolve } = require('path');
const request = require('request');
const archive = require('./archive');
const dw = require(process.cwd() + '/dw.json');
const hostname = dw.ocapi_endpoint ? dw.ocapi_endpoint : dw.hostname;

const migrationsRootDir = process.cwd() + '/migrations/';

const getCurrentMigration = (token) => {
    return new Promise((resolve, reject) => {
        if (token) {
            
            const options = {
                url: 'https://' + hostname + '/s/-/dw/data/v21_3/global_preferences/preference_groups/migrations/current',
                method: 'GET',
                headers: {
                    'Authorization': 'Bearer ' + token
                }
            };

            request(options, (err, res, body) => {
                let json = JSON.parse(body);

                if ((json.fault && json.fault.type === "CustomPreferenceGroupNotFoundException") || !json.c_migrationVersion) {
                    resolve(false);
                } else {
                    const instanceType = json.link.split('/').pop();
                    resolve({ currentMigration: json.c_migrationVersion, instanceType: instanceType });
                }
            });
        } else {
            reject(false);
        }
    });
}

const getMigrationsToRun = (currentMigration) => {
    const getDirectories = source =>
        fs.readdirSync(source, { withFileTypes: true })
            .filter(dirent => dirent.isDirectory())
            .map(dirent => dirent.name);

    let migrations = getDirectories(migrationsRootDir).sort();
    
    if (currentMigration) {
        console.log('Current migration on instance: ' + currentMigration);

        const currentMigrationIndex = migrations.findIndex(element => currentMigration === element);

        if (currentMigrationIndex === -1) {
            console.log('Current migration (' + currentMigration + ') not found.');
            migrations = [];
            process.exitCode = 1;
        } else {
            migrations = migrations.slice(currentMigrationIndex + 1);
        }
    } else {
        console.log('No migration on instance. Running all migrations ...');
    }

    return migrations;
}

const addMigration = (description) => {
    let rightNow = new Date();
    let timeStamp = rightNow.toISOString().replace(/-|T|:|\.|Z/g,"");
    let descriptionSlug = description.toLowerCase().replace(/ /g,'-').replace(/[^\w-]+/g,'');

    let folderName = timeStamp + '_' + descriptionSlug;

    let directory = resolve(migrationsRootDir, folderName);

    if (!fs.existsSync(migrationsRootDir)) {
        fs.mkdirSync(migrationsRootDir);
    }

    if (!fs.existsSync(directory)){
        fs.mkdirSync(directory);
    }
};

const setMigration = (migrationId, token, instanceType) => {
    return new Promise((resolve, reject) => {
        if (token) {
            const options = {
                url: 'https://' + hostname + '/s/-/dw/data/v21_3/global_preferences/preference_groups/migrations/' + instanceType,
                method: 'PATCH',
                headers: {
                    'Authorization': 'Bearer ' + token
                },
                json: { c_migrationVersion: migrationId }
            };

            request(options, (err, res, body) => {
                resolve({success: true});
            });
        } else {
            reject({success: false});
        }
    });
};

const applyMigration = (token) => {
    getCurrentMigration(token).then(({currentMigration, instanceType}) => {

        let migrations = getMigrationsToRun(currentMigration);

        let archiveMigration = (filename, directory) => {
            return new Promise((resolve, reject) => {
                archive('migration', filename, [directory], token).then((res) => {
                    resolve();
                }).catch((err) => {
                    reject(err);
                });
            });
        };

        let pushMigrations = () => {
            return new Promise((resolve, reject) => {
                let pushMigration = (index) => {
                    if (index === migrations.length) {
                        resolve();
                        return;
                    }
            
                    let element = migrations[index];
                    let filename = element;
                    let directory = migrationsRootDir + element;
                    
                    archiveMigration(filename, directory).then(() => {
                        console.log('Migrating: ' + element);
                        index++;
                        currentMigration = element;

                        setMigration(currentMigration, token, instanceType).then(() => {
                            pushMigration(index);
                        }).catch((error) => {
                            console.log(error);
                        });
                    }).catch((err) => {
                        reject('Something went wrong with migration ' + element);
                        return;
                    });
                };

                pushMigration(0);
            });
        };

        if (migrations.length > 0) {
            pushMigrations().then((response) => {
                console.log('Migrations done!');
            }).catch((err) => {
                console.log(err);
            });
        } else {
            console.log('Migrations are up to date');
        }
    });
};

module.exports = {
    create: addMigration,
    set: setMigration,
    apply: applyMigration
};